#include <iostream>
#include <fstream>
#include <iomanip>
#include <string>
#include <vector>
#include <cmath>
#include <cstdlib>
#include <cctype>
#include <sstream>
#include <ctime>

//timeHorizon: 42, 56 or 72
//Instance		A		B		C		D		E		F
//----------------------------------------------------------
//instance		1		2		3		4		5		6
//numShips		1		2		2		2		2		3
//numPorts		6		5		6		7		7		6
//numPickup		1		1		2		2		2		2
//numDelivery	3		2		2		3		3		2

///////////////////////////
////TYPE INSTANCE HERE/////
#define instance 1
#define timeHorizon 42
#define numShips 1
#define numPorts 6
#define numPickup 1
#define numDelivery 3
//////////////////////////
//////////////////////////

using namespace std;

class Route{
public:
	int vesselName;
	double cost;
	int totalTime;
	vector<int> georoute;
	Route(int vesselNum){vesselName = vesselNum; cost = 0; totalTime = 1;}
	Route(){vesselName = 0; cost = 0; georoute.push_back(0);}
};

class Loading{
public:
	int nTanks;
	int port;
	int ship;
	int time;
	int route;
	Loading(){nTanks = 0; port = 0; ship = 0; time = 0, route = 0;}
};

class Operation{
public:
	int port;
	int time;
	Operation(int theport, int thetime){port = theport; time = thetime;}
	Operation(){port = 0; time = 0;}
};

class OperationRoute{
public:
	vector<Operation> operations;
	int route;
	int duplicate;
	double cost;
	int ship;
	OperationRoute(){operations.empty(); route = 0; duplicate = 1; cost = 0.0; ship = 0;}
};

class RouteCost{
public:
	int ship;
	double cost;
	RouteCost(int theship, double thecost){ship = theship; cost = thecost;}
	RouteCost(){ship = 0; cost = 0;}
};

void createRoute(const int &port, int tempPort,const int &currentShip, Route inroute, vector<Route> & shipsroutes, const int TimeIJV[][numPorts][numPorts], const double CostIJV[][numPorts][numPorts]);

void createPattern(vector<OperationRoute>::iterator it, Loading pd1, Loading pd2, Loading dd1, Loading dd2, int currentj, vector<OperationRoute> &alloperations, vector<Loading> pattern, vector<vector<Loading>> &allpatterns);

void createLoadingMatrices(vector<vector<Loading>> allpatterns, vector<Loading> &allloading, vector<Loading> &allunloading);

bool checkWaitingDays(Route inroute);

bool checkPickup(int port);

bool checkDelivery(int port);

int countDelivery(Route r);

bool checkDoubleDelivery(vector<Operation> operations, int j);

string convertInt(int number)
{
   stringstream ss;
   ss << number;
   return ss.str();
}

string printPort(int p){
	stringstream ss;
	if (p > 1 && p <= 1+numPickup){
		ss << "P" << p-1;}
	if (p > numPickup+1 && p < numPorts){
		ss << "D" << p-numPickup-1;}
	return ss.str();
}

string convertInstance(int inst)
{
	char instanceLetter = inst + 64;
	stringstream ss;
	ss << instanceLetter;  
   return ss.str();
}

int main(){
	
	//START TIMER
	clock_t start = clock();
	double diff;
	//---------------

	//LESE FRA FIL
	ifstream fin;
	fin.open("DataFile_"+convertInstance(instance)+".txt");
	if(fin.fail()){ cout << "Failed to open file"; exit(1);}
	else{cout << "File open";}

	string text; //midlertidig variabel som fjerner tekst fra filen
	fin >> text;

	//------------------------------
	//LESER INN PORTS FRA FIL
	vector <string> ports;
	string port;
	for(int i = 0; i < numPorts; i++){ 
		fin >> port; 
		ports.push_back(port);
	}
	//------------------------------

	cout << endl;
	fin >> text;
	//------------------------------
	//LESER INN TIDSMATRISE FRA FIL
	int TimeIJV[numShips][numPorts][numPorts];
		for(int i = 0; i < numShips; i++){
		for(int j = 0; j < numPorts; j++){
			for(int k = 0; k < numPorts; k++){
				fin >> TimeIJV[i][j][k];
			}
		}
	}
	//------------------------------

	fin >> text;
	//------------------------------
	//LESER INN KOSTNADSMATRISE FRA FIL
	double CostIJV[numShips][numPorts][numPorts];
	for(int i = 0; i < numShips; i++){
		for(int j = 0; j < numPorts; j++){
			for(int k = 0; k < numPorts; k++){
				fin >> CostIJV[i][j][k];
			}
		}
	}
	//------------------------------
	
	fin.close();

	//SKRIVER UT TIDS- OG KOSTNADSMATRISE FOR SJEKK
		for(int i = 0; i < numShips; i++){
		for(int j = 0; j < numPorts; j++){
			for(int k = 0; k < numPorts; k++){
				cout << TimeIJV[i][j][k] << " ";
			}
			cout << endl;
		}
		cout << endl;
	}
	cout << endl;
			for(int i = 0; i < numShips; i++){
		for(int j = 0; j < numPorts; j++){
			for(int k = 0; k < numPorts; k++){
				cout << CostIJV[i][j][k] << " ";
			}
			cout << endl;
		}
		cout << endl;
	}
	//----------------------------------------
	
	int currentPort = 1;
	int tempPort = 1;
	int currentShip = 1;
	vector<Route> shipsroutes;
	cout << "Creating routes! \n";
	for (int vv = 0; vv < numShips; vv++){
		Route myRoute(currentShip);
		createRoute(currentPort, tempPort, currentShip, myRoute, shipsroutes, TimeIJV, CostIJV);
		currentShip++;
	}
	cout << "Finished creating routes. \n";
	
	//Lage O-matrisen og Cost-matrisen
	int tempTime = 1;
	int tempCost = 0;
	vector<OperationRoute> alloperations;
	vector<RouteCost> allcosts;
	OperationRoute newOperationRoute;
	Operation o;
	int n = shipsroutes.size()-1;
	while (!shipsroutes.empty()){
		if(!checkPickup(shipsroutes[n].georoute[shipsroutes[n].georoute.size()-2])){ //Kan ikke avslutte i pickup
			newOperationRoute.operations.clear();
			tempTime = 1+ TimeIJV[shipsroutes[n].vesselName-1][(shipsroutes[n].georoute[0])-1][(shipsroutes[n].georoute[1])-1];			//Oppdatere tid (for operation)
			newOperationRoute.ship = shipsroutes[n].vesselName;		//bestemme skipet som seiler ruta
			newOperationRoute.cost = shipsroutes[n].cost;			//tilegne kostnad
				for (unsigned int j = 2; j < shipsroutes[n].georoute.size(); j++){
					if(shipsroutes[n].georoute[j-1] != shipsroutes[n].georoute[j] && shipsroutes[n].georoute.size() >= 2){				//Nr skipet forlater havn trigges O-parameteren
						o.port = shipsroutes[n].georoute[j-1];		//port for operation
						o.time = tempTime;							//tid for operation
						newOperationRoute.operations.push_back(o);	//legg til siste operation i gjeldende sekvens
					}
					tempTime += TimeIJV[shipsroutes[n].vesselName-1][(shipsroutes[n].georoute[j-1])-1][(shipsroutes[n].georoute[j])-1];	 //Oppdatere tid (for operation)
				}
			if(!newOperationRoute.operations.empty()){
				alloperations.push_back(newOperationRoute);			//legg til gjeldende sekvens av operations i lista over sekvenser
			}
		}
		shipsroutes.pop_back();
		n--;
	}


	//Lage laste/losse-mnster
	cout << "Creating patterns!\n";
	Loading p1, p2, d1, d2;
	int currentj = 0;
	vector<Loading> pattern;
	vector< vector<Loading> > allpatterns;
	vector<OperationRoute>::iterator it = alloperations.begin();
	while(it != alloperations.end()){
		createPattern(it, p1, p2, d1, d2, currentj, alloperations, pattern, allpatterns);
		it++;
	}
	cout << "Finished creating patterns.\n" ;
	diff = (clock() - start ) / (double)CLOCKS_PER_SEC;
	cout << "Time to generate all routes: " << diff << " seconds\n" << endl;
	

	//SKRIV TIL FIL
	cout << "Writes to file!\n";
	ofstream outputFile;
	outputFile.open("DataFile_"+convertInstance(instance)+"_4_" + convertInt(timeHorizon) + "_"+convertInt(timeHorizon/2)+"_path.txt");
	//outputFile.open("DataFile_"+convertInstance(instance)+"_4_" + convertInt(timeHorizon) + "_"+convertInt(timeHorizon/2)+"_1day.txt"); //max one waiting day outside port
	if (outputFile.fail()){
		cout << "Output file opening failed. \n";
		exit(1);
	}

	outputFile << "!Time to generate all routes: " << diff << " seconds\n" << endl;

	int nRoutes = 0;
	int corr = 0;
	
	outputFile << endl << "Nallroutes: " << allpatterns.size()+numShips <<endl ;
	
	outputFile << "PortOperations: [ \n";
	for (unsigned int i = 0; i < alloperations.size(); i++){
		for (unsigned int k = 0; k < alloperations[i].duplicate; k++){
			nRoutes++;
			for (unsigned int j = 0; j < alloperations[i].operations.size(); j++){
					outputFile << "('S" << alloperations[i].ship << "', '" << printPort(alloperations[i].operations[j].port) << "', " << alloperations[i].operations[j].time << ", " << nRoutes << ") " << 1 << "\n";
			}
		}
	}
	outputFile << "]\n" << endl;
	cout << "PortOperations written\n";
	
	nRoutes = 0;
	corr = 0;
	outputFile << "Loading: [ \n";
	for (unsigned int i = 0; i < allpatterns.size(); i++){
		nRoutes++;
		for (unsigned int j = 0; j < allpatterns[i].size(); j++){
			if(allpatterns[i][j].nTanks > 0){
				outputFile << "('S" << allpatterns[i][j].ship << "', '" << printPort(allpatterns[i][j].port) << "', " << allpatterns[i][j].time << ", " << nRoutes << ") " << allpatterns[i][j].nTanks << "\n";
			}
		}
	}
	outputFile << "]\n" << endl;
	cout << "Loading written\n";

	nRoutes = 0;
	outputFile << "Unloading: [ \n";
	for (unsigned int i = 0; i < allpatterns.size(); i++){
		nRoutes++;
		for (unsigned int j = 0; j < allpatterns[i].size(); j++){
			if(allpatterns[i][j].nTanks < 0){
				outputFile << "('S" << allpatterns[i][j].ship << "', '" << printPort(allpatterns[i][j].port) << "', " << allpatterns[i][j].time << ", " << nRoutes << ") " << -allpatterns[i][j].nTanks << "\n";
			}
		}
	}
	outputFile << "]\n" << endl;
	cout << "Unloading written\n";

	nRoutes = 0;
	outputFile << "COSTSailing: [ \n";
	for (unsigned int i = 0; i < alloperations.size(); i++){
		for (unsigned int k = 0; k < alloperations[i].duplicate; k++){
			nRoutes++;
			outputFile << "('S" << alloperations[i].ship << "', " << nRoutes << ") " << alloperations[i].cost << endl;
		}
	}
	outputFile << "]\n" << endl;
	cout << "Costs written\n";
		
	outputFile.close();
	
	//STOP TIMER AND PRINT TIME	
	diff = (clock() - start ) / (double)CLOCKS_PER_SEC;
	cout << "Total time: " << diff << " seconds\n" << endl;
	system("pause");
	//-------------------------
	return 0;
}


void createRoute(const int &port, int tempPort, const int &currentShip, Route inroute, vector<Route> & shipsroutes, const int TimeIJV[][numPorts][numPorts], const double CostIJV[][numPorts][numPorts]){
		inroute.georoute.push_back(port);									 //legg til port i gjeldende rute
		int deliveryInRow;
		if(inroute.georoute.size() > 1){									 //dersom flere enn n port i ruta
			inroute.totalTime += TimeIJV[currentShip-1][tempPort-1][port-1]; //oppdatere tid ruta tar
			inroute.cost += CostIJV[currentShip-1][tempPort-1][port-1];		//oppdatere kostnad for ruta
			deliveryInRow = countDelivery(inroute);							 //sjekk antall delivery ports som beskes p rad
			
		}
		bool checkWaiting = false;
		if(inroute.georoute.size() > 2){
		checkWaiting = checkWaitingDays(inroute);	 //sjekk antall sammenhengende ventedager p dette tidspunktet
		}
		if(inroute.totalTime <= timeHorizon){		//srg for at ruta ikke er lengre enn planleggingshorisonten
			for (int jj = 2; jj <= numPorts; jj++){	//srg for at rutene utgjr alle kombinasjoner av port calls
				if (TimeIJV[currentShip-1][port-1][jj-1] != -1 && (inroute.totalTime + TimeIJV[currentShip-1][port-1][jj-1]) <= timeHorizon)// && port != jj){ //sjekk at arc (i,j) kan seiles
					if (jj != numPorts){ //not destination port
						if((checkWaiting && (port != jj)) || !checkWaiting){ //hvis antall ventedager har ndd grensen og neste havn ikke er destination port, eller antall ventedager er under grensen
							if( (checkPickup(port) && checkDelivery(jj)) || port == 1 || port == jj || ((deliveryInRow==1 || deliveryInRow==2) && checkPickup(jj)) || (checkDelivery(port) && deliveryInRow < 2)){ //routing-krav
								tempPort = port;
								createRoute(jj, tempPort, currentShip, inroute, shipsroutes, TimeIJV, CostIJV);  //rekursivt kall, srger for ruter der alle ports blir kombinert
							}
						}
					}
					else{
					inroute.georoute.push_back(numPorts);	//legg til destination node, ruta avsluttes
					shipsroutes.push_back(inroute);			//legg til ruta i listen av mulige ruter
					}
				}
		}
		else{
			inroute.georoute.push_back(numPorts);			 //legg til destination node, ruta avsluttes
			shipsroutes.push_back(inroute);					 //legg til ruta i listen av mulige ruter
		}
}

//Skaper alle mulige laste/losse-mnster
void createPattern(vector<OperationRoute>::iterator it, Loading pd1, Loading pd2, Loading dd1, Loading dd2, int currentj, vector<OperationRoute> &alloperations, vector<Loading> pattern, vector<vector<Loading>> &allpatterns){
	if (currentj > 0){
		pattern.push_back(pd1);
		pattern.push_back(pd2);
		pattern.push_back(dd1);
		pattern.push_back(dd2);
	}
	Loading p, d;
	while(currentj < it->operations.size()-1 && it->operations.size() >= 1){
		if(checkPickup(it->operations[currentj].port)){ //hvis havnen er pick-up havn
			if(!checkDoubleDelivery(it->operations, currentj)){  //hvis bare n delivery ports beskes i den aktuelle duty
				p.nTanks = 4;								//last fire tanker
				p.port = it->operations[currentj+1].port;	//til neste port
				p.ship = it->ship;							//med gjeldende skip
				p.time = it->operations[currentj].time;		//i den aktuelle tidsperioden
				d.nTanks = -4;								//losse fire tanker
				d.port = it->operations[currentj+1].port;	//til neste port
				d.ship = it->ship;							//med gjeldende skip
				d.time = it->operations[currentj+1].time;	//nr neste port beskes
				pattern.push_back(p);	//legg til lasting i laste/losse-mnster
				pattern.push_back(d);	//legg til lossing i laste/losse-mnster
			}
				else if (checkDoubleDelivery(it->operations, currentj)){  //hvis to delivery ports beskes i den aktuelle duty
					it->duplicate++;		//dupliser den aktuelle raden (i portoperations-matrisen)
					Loading pd1, pd2, dd1, dd2;
					for (int i = 2; i > 1; i--){					//for-lkka er undvendig nr bare [1,3]- og [3,1]-mnsteret genereres. ndvendig for  generere [1,3]+[2,2]+[3,1]
						pd1.nTanks = 1+i;							//last 3 tank
						pd1.port = it->operations[currentj+1].port;	//til neste port
						pd1.ship = it->ship;						
						pd1.time = it->operations[currentj].time;
						pd2.nTanks = 3-i;							//last 1 tanker
						pd2.port = it->operations[currentj+2].port;	//til siste delivery port i duty
						pd2.ship = it->ship;
						pd2.time = it->operations[currentj].time;
						dd1.nTanks = -(1+i);						//losse 1 tank
						dd1.port = it->operations[currentj+1].port;	//til neste port
						dd1.ship = it->ship;
						dd1.time = it->operations[currentj+1].time;
						dd2.nTanks = -(3-i);						//losse 3 tanker
						dd2.port = it->operations[currentj+2].port;	//til siste delivery port i duyt
						dd2.ship = it->ship;
						dd2.time = it->operations[currentj+2].time;
						createPattern (it, pd1, pd2, dd1, dd2, currentj+1, alloperations, pattern, allpatterns); //rekursivt kall. srger for  genere alle laste/losse-mnstre
					}
						pd1.nTanks = 1;								//last 1 tank
						pd1.port = it->operations[currentj+1].port;	//til neste port
						pd1.ship = it->ship;
						pd1.time = it->operations[currentj].time;
						pd2.nTanks = 3;								//last 3 tanker
						pd2.port = it->operations[currentj+2].port;	//til siste delivery port i duty
						pd2.ship = it->ship;
						pd2.time = it->operations[currentj].time;
						dd1.nTanks = -1;							//losse 1..osv.
						dd1.port = it->operations[currentj+1].port;
						dd1.ship = it->ship;
						dd1.time = it->operations[currentj+1].time;
						dd2.nTanks = -3;
						dd2.port = it->operations[currentj+2].port;
						dd2.ship = it->ship;
						dd2.time = it->operations[currentj+2].time;
						pattern.push_back(pd1);
						pattern.push_back(pd2);
						pattern.push_back(dd1);
						pattern.push_back(dd2);
				}
		//}
		}
		currentj++;
	}
	if(!pattern.empty()){
	allpatterns.push_back(pattern); //legg til laste/losse-mnster i listen over alle mnstre
	}
}

//Max two waiting days
bool checkWaitingDays(Route inroute){
	return (inroute.georoute[inroute.georoute.size()-1] == inroute.georoute[inroute.georoute.size()-2] && inroute.georoute[inroute.georoute.size()-1] == inroute.georoute[inroute.georoute.size()-3]);
}

//Max one waiting day
//bool checkWaitingDays(Route inroute){
//	return (inroute.georoute[inroute.georoute.size()-1] == inroute.georoute[inroute.georoute.size()-2]);
//}

bool checkDoubleDelivery(vector<Operation> operations, int j){
	if ( j < operations.size()-2 ){
		if (checkDelivery(operations[j+1].port) && checkDelivery(operations[j+2].port)){return true;}
		else {return false;}
	}
	else {return false;}
}

bool checkPickup(int port){
	return (port > 1 && port < numPickup+2);
}

bool checkDelivery(int port){
	return (port > 1+numPickup && port < numPorts);
}

int countDelivery(Route r){
	int count = 0;
	for (unsigned int i = 1; i < r.georoute.size(); i++){
		if(checkDelivery(r.georoute[i]) && (r.georoute[i] != r.georoute[i-1])){count++;}
		if(checkPickup(r.georoute[i])){count = 0;}
	}
	return count;
}

//LAGER LOADINGMATRISEN LIKE STOR SOM PATTERNS
void createLoadingMatrices(vector<vector<Loading>> allpatterns, vector<Loading> &allloading, vector<Loading> &allunloading){
	int nRoutes = 1;
	int check = 1;
	for (unsigned int i = 0; i < allpatterns.size(); i++){
		if(check != allpatterns[i][0].ship){
			check++;
			nRoutes++;
		}
			for (int j = 0; j < allpatterns[i].size(); j++){
				allpatterns[i][j].route = nRoutes;
				if(allpatterns[i][j].nTanks > 0){
					allloading.push_back(allpatterns[i][j]);
				}
				else{
					allpatterns[i][j].nTanks = -allpatterns[i][j].nTanks;
					allunloading.push_back(allpatterns[i][j]);
				}
			}
	nRoutes++;
	}
}
